from werkzeug.routing import Rule
from werkzeug.wrappers import Response, request
from functools import wraps
from werkzeug.exceptions import HTTPException, MethodNotAllowed

class Blueprint:
    def __init__(self, name, import_name, url_prefix=''):
        self.name = name
        self.import_name = import_name
        self.url_prefix = url_prefix
        self.url_map = []

    def route(self, rule, **options):
        def decorator(f):
            endpoint = options.pop('endpoint', f.__name__)
            methods = options.pop('methods', ['GET'])
            secure = options.pop('secure', False)
            self.url_map.append((rule, endpoint, f, methods, secure))
            return f
        return decorator

    def register(self, zylo_app):
        for rule, endpoint, f, methods, secure in self.url_map:
            full_rule = f'{self.url_prefix}{rule}'
            zylo_app.url_map.add(Rule(full_rule, endpoint=endpoint))
            zylo_app.routes[endpoint] = (full_rule, methods, f)

            if secure:
                zylo_app.before_request_funcs.append(self._enforce_https(f))

    def _enforce_https(self, f):
        @wraps(f)
        def decorator(*args, **kwargs):
            if not request.is_secure:
                response = Response()
                response.status_code = 301
                response.headers['Location'] = request.url.replace('http://', 'https://')
                return response
            return f(*args, **kwargs)
        return decorator


def blueprint_route(rule, **options):
    def decorator(f):
        endpoint = options.pop('endpoint', f.__name__)
        f.blueprint_rule = (rule, endpoint)
        return f
    return decorator

def register_blueprint(app, blueprint):
    blueprint.register(app)



import hashlib
import secrets
import base64
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import os
import random
import string
import re
from html.parser import HTMLParser

SALT_LENGTH = 16
PEPPER_LENGTH = 16

def generate_salt():
    return secrets.token_bytes(SALT_LENGTH)

def generate_pepper():
    return secrets.token_bytes(PEPPER_LENGTH)

def hash_password(password):
    salt = generate_salt()
    pepper = b'$zylo.chiper@'
    combined = salt + password.encode() + pepper

    for _ in range(20):
        combined = hashlib.sha3_512(combined).digest()
        combined = hashlib.sha3_256(combined).digest()

    iterations = 2000000
    for _ in range(iterations):
        combined = hashlib.sha3_512(combined).digest()

    final_hash = '$zylo.chiper@' + salt.hex() + combined.hex()
    return final_hash

def verify_password(password, hashed_password):
    prefix = '$zylo.chiper@'
    salt_hex = hashed_password[len(prefix): len(prefix) + 32]
    combined_hash = hashed_password[len(prefix) + 32:]
    combined = bytes.fromhex(salt_hex) + password.encode() + prefix.encode()

    for _ in range(20):
        combined = hashlib.sha3_512(combined).digest()
        combined = hashlib.sha3_256(combined).digest()

    iterations = 2000000
    for _ in range(iterations):
        combined = hashlib.sha3_512(combined).digest()

    combined_hex = combined.hex()
    generated_hash = prefix + salt_hex + combined_hex
    return generated_hash == hashed_password

def encrypt_data(password, plaintext):
    try:
        salt = os.urandom(16)
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=100000,
        )
        key = base64.urlsafe_b64encode(kdf.derive(password.encode()))
        cipher_suite = Fernet(key)

        cipher_text = cipher_suite.encrypt(plaintext.encode())
        return salt + cipher_text

    except Exception as e:
        print("Encryption error:", str(e))


def decrypt_data(password, ciphertext):
    try:
        salt = ciphertext[:16]
        cipher_text = ciphertext[16:]
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=100000,
        )
        key = base64.urlsafe_b64encode(kdf.derive(password.encode()))
        cipher_suite = Fernet(key)

        decrypted_text = cipher_suite.decrypt(cipher_text)
        return decrypted_text.decode()

    except Exception as e:
        print("Decryption error:", str(e))


def generate_id(length=10):
    if length < 2:
        raise ValueError("Length must be at least 2.")
    starting_char = random.choice(string.ascii_lowercase)
    remaining_length = length - 2
    random_digits = ''.join(random.choice(string.digits) for _ in range(remaining_length))
    unique_id = starting_char + '-' + random_digits
    return unique_id

def validate_email(email):
    email_regex = r"[^@]+@[^@]+\.[^@]+"
    return re.match(email_regex, email) is not None

def is_secure_password(password):
    if len(password) < 8:
        return False

    if not any(char.isupper() for char in password):
        return False
    if not any(char.islower() for char in password):
        return False
    if not any(char.isdigit() for char in password):
        return False
    if not any(char in "!@#$%^&*()-_=+{}[]|;:,<.>/?`~" for char in password):
        return False

    return True

def chiper_key(node=None, clock_seq=None):
    from time import time
    import random

    nanoseconds = int(time() * 1e9)
    timestamp = int(nanoseconds / 100) + 0x01b21dd213814000
    clock_seq = clock_seq if clock_seq is not None else random.getrandbits(14)

    timestamp_bytes = timestamp.to_bytes(8, 'big')
    clock_seq_bytes = clock_seq.to_bytes(2, 'big')
    node = node.to_bytes(10, 'big') if node is not None else os.urandom(10)

    return b"".join([timestamp_bytes[:6], timestamp_bytes[6:], clock_seq_bytes, node])

def chiper_encode(namespace, name):
    namespace_bytes = namespace if isinstance(namespace, bytes) else namespace.encode()
    name_bytes = name.encode() if isinstance(name, str) else name

    hash_value = hashlib.md5(namespace_bytes + name_bytes).digest()
    hash_value = hash_value[:16] 

    return bytes(hash_value)

class SanitizeHTMLParser(HTMLParser):
    def __init__(self, allowed_tags=None, allowed_attributes=None):
        super().__init__()
        self.allowed_tags = allowed_tags or []
        self.allowed_attributes = allowed_attributes or []
        self.sanitized_data = []

    def handle_starttag(self, tag, attrs):
        if tag in self.allowed_tags:
            sanitized_attrs = [(attr, self.sanitize_sql_injection(value)) for attr, value in attrs if attr in self.allowed_attributes]
            sanitized_starttag = "<" + tag + " " + " ".join([attr + '="' + value + '"' for attr, value in sanitized_attrs]) + ">"
            self.sanitized_data.append(sanitized_starttag)

    def handle_endtag(self, tag):
        if tag in self.allowed_tags:
            self.sanitized_data.append("</" + tag + ">")

    def handle_data(self, data):
        self.sanitized_data.append(data)

    @staticmethod
    def sanitize_sql_injection(value):
        return re.sub(r"[\'\";]", '', value)


def sanitize_input(input_string, sanitize_html=True, sanitize_sql=True, sanitize_nosql=True):
    if sanitize_html:
        input_string = sanitize_html_input(input_string)

    if sanitize_sql:
        input_string = sanitize_sql_input(input_string)

    if sanitize_nosql:
        input_string = sanitize_nosql_input(input_string)

    return input_string


def sanitize_html_input(input_string):
    input_string = re.sub(r'<script.*?</script>', '', input_string, flags=re.DOTALL)
    input_string = re.sub(r'<style.*?</style>', '', input_string, flags=re.DOTALL)

    parser = SanitizeHTMLParser(
        allowed_tags=['p', 'br', 'strong', 'em', 'u'],
        allowed_attributes=['href', 'title']
    )
    parser.feed(input_string)

    return ''.join(parser.sanitized_data)


def sanitize_sql_input(input_string):
    return re.sub(r"[\'\";]", '', input_string)


def sanitize_nosql_input(input_string):
    return input_string.replace('$', '').replace('.', '')



import secrets
from datetime import datetime, timedelta

class SessionManager:
    def __init__(self):
        self.sessions = {}
        self.session_duration = timedelta(minutes=30)

    def generate_session_token(self):
        return secrets.token_hex(16)

    def create_session(self):
        token = self.generate_session_token()
        expires_at = datetime.now() + self.session_duration
        self.sessions[token] = {'expires_at': expires_at, 'data': {}}
        return token

    def get_session(self, token):
        session = self.sessions.get(token)
        if session and session['expires_at'] > datetime.now():
            return session['data']
        return {}

    def update_session(self, token, data):
        if token in self.sessions:
            session = self.sessions[token]
            if session['expires_at'] > datetime.now():
                session['data'] = data
                session['expires_at'] = datetime.now() + self.session_duration
                return session['data']
        return {}

    def delete_session(self, token):
        if token in self.sessions:
            del self.sessions[token]
            return True
        return False



import json
from werkzeug.wrappers import Request, Response
from werkzeug.routing import Map, Rule
from werkzeug.exceptions import HTTPException, default_exceptions, NotFound, MethodNotAllowed
from jinja2 import Environment, FileSystemLoader
from .sessions import SessionManager
from werkzeug.urls import url_encode
from .limiter import Limiter
import os
import mimetypes

class Zylo:
    def __init__(self, __name__={} ):
        self.url_map = Map()
        self.routes = {}
        self.template_env = Environment(loader=FileSystemLoader('views'))
        self.static_folder = 'static'
        self.session_manager = SessionManager()
        self.blueprints = []
        self.template_folder = None
        self.limiter = Limiter()
        self.before_request_funcs = []
        self.after_request_funcs = []
        self.__name__ = __name__

    def route(self, rule, methods=['GET'], endpoint=None,):
        def decorator(func):
            self.routes[endpoint or func.__name__] = (rule, methods, func)
            return func
        return decorator

    def register_blueprint(self, blueprint):
        self.blueprints.append(blueprint)
        blueprint.register(self)
        
    def url_for_static(self, filename):
        return f'/static/{filename}'
    
    def serve_static(self, filename):
        static_path = os.path.join(self.static_folder, filename)
        if os.path.isfile(static_path):
            mimetype, _ = mimetypes.guess_type(static_path)
            if mimetype:
                return Response(open(static_path, 'rb').read(), mimetype=mimetype)
        raise NotFound()
    
    def before_request(self, func):
        self.before_request_funcs.append(func)
        return func

    def after_request(self, func):
        self.after_request_funcs.append(func)
        return func

    def run(self, host='localhost', port=8000, debug=True):
        app = self
        host = host
        port = port
        debug = debug

        @Request.application
        def application(request):
            try:
                session_token = request.cookies.get('session_token')
                session_data = app.session_manager.get_session(session_token)
                request.session = session_data

                adapter = app.url_map.bind_to_environ(request.environ)
                try:
                    endpoint, values = adapter.match()
                except NotFound:
                    for blueprint in app.blueprints:
                        adapter = blueprint.url_map.bind_to_environ(request.environ)
                        try:
                            endpoint, values = adapter.match()
                            break
                        except NotFound:
                            pass
                    else:
                        raise
                rule, methods, func = app.routes[endpoint]
                if request.method not in methods:
                    raise MethodNotAllowed(valid_methods=methods)

                if hasattr(request, 'session'):
                    if not session_token or not session_data:
                        session_token = app.session_manager.create_session()
                        session_data = {}
                        request.session = session_data
                        response = func(request, **values)
                        response.set_cookie('session_token', session_token)
                    else:
                        response = func(request, **values)

                    app.session_manager.update_session(session_token, request.session)

                return response

            except HTTPException as e:
                return e

        if debug:
            from werkzeug.debug import DebuggedApplication
            application = DebuggedApplication(application, evalex=True)

        for endpoint, (rule, _, _) in self.routes.items():
            self.url_map.add(Rule(rule, endpoint=endpoint))
            print(f"Registered route: {rule} --> {endpoint}")

        for blueprint in self.blueprints:
            blueprint.register(self)

        from werkzeug.serving import run_simple
        run_simple(host, port, application, use_reloader=debug)

app = Zylo()

def render_template(template_name, **kwargs):
    template = app.template_env.get_template(template_name)
    kwargs['url_for_static'] = app.url_for_static
    return Response(template.render(**kwargs), mimetype='text/html')

def jsonify(data):
    response = Response()
    response.headers['Content-Type'] = 'application/json'
    response.data = json.dumps(data)
    return response

def redirect(location):
    response = Response()
    response.status_code = 302
    response.headers['Location'] = location
    return response

def url_for(endpoint):
    rule, _, _ = app.routes[endpoint]
    url_map = app.url_map.bind(app.host, app.port)
    return url_map.build(rule, force_external=True)

def redirect_args(location, **kwargs):
    url = location
    if kwargs:
        query_params = url_encode(kwargs)
        url += f'?{query_params}'
    return Response(status=302, headers={'Location': url})

def send_file(filename, mimetype=None, as_attachment=False):
    response = Response()
    response.headers['Content-Disposition'] = f'attachment; filename="{filename}"' if as_attachment else f'inline; filename="{filename}"'
    if mimetype is not None:
        response.headers['Content-Type'] = mimetype
    return response

def static_engine(static_folder):
    app.static_folder = static_folder


def template_engine(template_folder):
    app.template_folder = template_folder
    app.template_env = Environment(loader=FileSystemLoader(template_folder))

# Add Werkzeug's default error handlers
for code in default_exceptions:
    handler = default_exceptions[code]

    def error_handler(error):
        response = Response()
        response.status_code = error.code
        response.data = f'{error.code} {error.name}'
        return response

    handler = error_handler




from werkzeug.wrappers import Response, request
from functools import wraps
from datetime import datetime, timedelta

class Limiter:
    def __init__(self, app=None):
        self.app = app
        self.cache = {}
        self.default_limits = []
        self.headers_enabled = True

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.limiter = self
        app.before_request(self.before_request)
        if self.headers_enabled:
            app.after_request(self.after_request)

    def before_request(self):
        rule = self.get_rule(request.url_rule)
        if rule is None:
            return None

        key = self.get_storage_key(rule.endpoint)
        limit = self.get_view_rate_limit(rule)

        if self.is_rate_limited(key, limit):
            return self.handle_rate_limit_exceeded(limit)

        self.increment_counter(key, limit)

    def after_request(self, response):
        response.headers.add('X-RateLimit-Limit', str(response.limit))
        response.headers.add('X-RateLimit-Remaining', str(response.remaining))
        reset_time = datetime.utcnow() + timedelta(seconds=response.reset)
        response.headers.add('X-RateLimit-Reset', reset_time.strftime('%Y-%m-%d %H:%M:%S'))
        return response

    def get_storage_key(self, endpoint):
        return f'limiter:{endpoint}:{request.remote_addr}'

    def get_view_rate_limit(self, rule):
        endpoint = rule.endpoint
        for limit, endpoint_func in self.default_limits:
            if endpoint_func is None or endpoint_func == endpoint:
                return limit

    def is_rate_limited(self, key, limit):
        counter = self.cache.get(key, 0)
        return counter >= limit

    def handle_rate_limit_exceeded(self, limit):
        response = Response()
        response.status_code = 429
        response.data = f'Rate limit exceeded. Limit: {limit}'
        return response

    def increment_counter(self, key, limit):
        self.cache.setdefault(key, 0)
        self.cache[key] += 1

    def limit(self, limit_value, key_func=None):
        def decorator(f):
            self.default_limits.append((limit_value, f.__name__))
            @wraps(f)
            def wrapped(*args, **kwargs):
                return f(*args, **kwargs)
            return wrapped
        return decorator



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='zylo',
    version='1.0.10',
    description='A lightweight web framework made with love',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Pawan kumar',
    author_email='control@vvfin.in',
    url='https://github.com/E491K7/zylo',
    packages=find_packages(),
    install_requires=['werkzeug', 'jinja2', 'cryptography', 'zylo-admin'],  
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)
